import json
import tarfile

from prefect import task


@task
def extract_tick_json(tar: tarfile.TarFile, key: str, tick: int, extension: str) -> list[dict]:
    member = tar.extractfile(f"{key}_{tick:06d}.{extension}.json")
    assert member is not None
    tick_json = json.loads(member.read().decode("utf-8"))
    return tick_json



import tarfile

import numpy as np
from prefect import task

from arcade_collection.output.extract_tick_json import extract_tick_json
from arcade_collection.output.get_location_voxels import get_location_voxels


@task
def convert_to_images(
    series_key: str,
    data_tar: tarfile.TarFile,
    frame_spec: tuple[int, int, int],
    regions: list[str],
    box: tuple[int, int, int],
    chunk_size: int,
    binary: bool,
) -> list[tuple[int, int, np.ndarray]]:
    length, width, height = box
    frames = list(np.arange(*frame_spec))
    array = np.zeros((len(frames), len(regions), height, width, length), "uint16")

    for index, frame in enumerate(frames):
        locations = extract_tick_json.fn(data_tar, series_key, frame, "LOCATIONS")

        for location in locations:
            location_id = location["id"]

            for channel, region in enumerate(regions):
                voxels = [
                    (z, y, x)
                    for x, y, z in get_location_voxels.fn(
                        location, region if region != "DEFAULT" else None
                    )
                ]
                array[index, channel][tuple(np.transpose(voxels))] = 1 if binary else location_id

    return split_array_chunks(array, chunk_size)


def split_array_chunks(array: np.ndarray, chunk_size: int) -> list[tuple[int, int, np.ndarray]]:
    chunks = []
    length = array.shape[4]
    width = array.shape[3]

    # Calculate chunk splits.
    length_section = (
        [0, chunk_size + 1] + (int(length / chunk_size) - 2) * [chunk_size] + [chunk_size + 1]
    )
    length_splits = np.array(length_section, dtype=np.int32).cumsum()
    width_section = (
        [0, chunk_size + 1] + (int(width / chunk_size) - 2) * [chunk_size] + [chunk_size + 1]
    )
    width_splits = np.array(width_section, dtype=np.int32).cumsum()

    # Iterate through each chunk split.
    for i in range(len(length_splits) - 1):
        length_start = length_splits[i]
        length_end = length_splits[i + 1]

        for j in range(len(width_splits) - 1):
            width_start = width_splits[j]
            width_end = width_splits[j + 1]

            # Extract chunk from full contents.
            chunk = array[:, :, :, length_start:length_end, width_start:width_end]

            if np.sum(chunk) != 0:
                chunks.append((i, j, chunk))

    return chunks



import tarfile

import numpy as np
from prefect import task
from simulariumio import (
    DISPLAY_TYPE,
    AgentData,
    CameraData,
    DimensionData,
    DisplayData,
    MetaData,
    ModelMetaData,
    TrajectoryConverter,
    TrajectoryData,
    UnitData,
)

from arcade_collection.output.extract_tick_json import extract_tick_json


@task
def convert_to_simularium(
    series_key: str,
    cells_data_tar: tarfile.TarFile,
    locations_data_tar: tarfile.TarFile,
    frame_spec: tuple[int, int, int],
    box: tuple[int, int, int],
    ds: float,
    dt: float,
    phase_colors: dict[str, str],
) -> str:
    length, width, height = box
    frames = list(np.arange(*frame_spec))

    meta_data = get_meta_data(series_key, length, width, height, ds)
    agent_data = get_agent_data(series_key, cells_data_tar, frames)
    agent_data.display_data = get_display_data(phase_colors)

    for index, frame in enumerate(frames):
        agent_data.times[index] = float(frame) * dt
        convert_cells_tar(agent_data, cells_data_tar, series_key, frame, index)
        convert_locations_tar(
            agent_data, locations_data_tar, series_key, frame, index, length, width, height, ds
        )

    return TrajectoryConverter(
        TrajectoryData(
            meta_data=meta_data,
            agent_data=agent_data,
            time_units=UnitData("hr"),
            spatial_units=UnitData("um"),
        )
    ).to_JSON()


def get_meta_data(series_key: str, length: int, width: int, height: int, ds: float) -> MetaData:
    meta_data = MetaData(
        box_size=np.array([length * ds, width * ds, height * ds]),
        camera_defaults=CameraData(
            position=np.array([10.0, 0.0, 200.0]),
            look_at_position=np.array([10.0, 0.0, 0.0]),
            fov_degrees=60.0,
        ),
        trajectory_title=f"ARCADE - {series_key}",
        model_meta_data=ModelMetaData(
            title="ARCADE",
            version="3.0",
            description=(f"Agent-based modeling framework ARCADE for {series_key}."),
        ),
    )

    return meta_data


def get_agent_data(series_key: str, cells_tar: tarfile.TarFile, frames: list[int]) -> AgentData:
    total_frames = len(frames)

    max_agents = 0
    for frame in frames:
        cells = extract_tick_json.fn(cells_tar, series_key, frame, "CELLS")
        max_agents = max(max_agents, len(cells))

    return AgentData.from_dimensions(DimensionData(total_frames, max_agents))


def get_display_data(phase_colors: dict[str, str]) -> DisplayData:
    display_data = {}

    for phase, color in phase_colors.items():
        display_data[phase] = DisplayData(name=phase, color=color, display_type=DISPLAY_TYPE.SPHERE)

    return display_data


def convert_cells_tar(
    data: AgentData,
    tar: tarfile.TarFile,
    series_key: str,
    frame: int,
    index: int,
) -> None:
    cells = extract_tick_json.fn(tar, series_key, frame, "CELLS")
    data.n_agents[index] = len(cells)

    for i, cell in enumerate(cells):
        data.unique_ids[index][i] = cell["id"]
        data.types[index].append(cell["phase"])
        data.radii[index][i] = (cell["voxels"] ** (1.0 / 3)) / 1.5


def convert_locations_tar(
    data: AgentData,
    tar: tarfile.TarFile,
    series_key: str,
    frame: int,
    index: int,
    length: int,
    width: int,
    height: int,
    ds: float,
) -> None:
    locations = extract_tick_json.fn(tar, series_key, frame, "LOCATIONS")

    for i, location in enumerate(locations):
        data.positions[index][i] = np.array(
            [
                (location["center"][0] - length / 2) * ds,
                (width / 2 - location["center"][1]) * ds,
                (location["center"][2] - height / 2) * ds,
            ]
        )



import tarfile

import numpy as np
from prefect import task
from skimage import measure

from arcade_collection.output.extract_tick_json import extract_tick_json
from arcade_collection.output.get_location_voxels import get_location_voxels


@task
def convert_to_meshes(
    series_key: str, data_tar: tarfile.TarFile, frame_spec: tuple[int, int, int], regions: list[str]
) -> list[tuple[int, int, str, str]]:
    frames = list(np.arange(*frame_spec))
    meshes = []

    for frame in frames:
        locations = extract_tick_json.fn(data_tar, series_key, frame, "LOCATIONS")

        for location in locations:
            location_id = location["id"]

            for region in regions:
                voxels = [
                    (z, y, x)
                    for x, y, z in get_location_voxels.fn(
                        location, region if region != "DEFAULT" else None
                    )
                ]
                array = make_mesh_array(voxels)
                mesh = make_mesh_object(array)

                meshes.append((frame, location_id, region, mesh))

    return meshes


def make_mesh_array(voxels: list[tuple[int, int, int]]) -> np.ndarray:
    mins = np.min(voxels, axis=0)
    maxs = np.max(voxels, axis=0)
    height, width, length = np.subtract(maxs, mins) + 3
    array = np.zeros((height, width, length), dtype=np.uint8)

    voxels_transposed = [voxel - mins + 1 for voxel in voxels]
    array[tuple(np.transpose(voxels_transposed))] = 7

    # Get set of zero neighbors for all voxels.
    offsets = [(-1, 0, 0), (1, 0, 0), (0, -1, 0), (0, 1, 0), (0, 0, -1), (0, 0, 1)]
    neighbors = {
        (z + i, y + j, x + k)
        for z, y, x in voxels_transposed
        for k, j, i in offsets
        if array[z + k, y + j, x + i] == 0
    }

    # Remove invalid neighbors on borders.
    neighbors = {
        (z, y, x)
        for z, y, x in neighbors
        if x != 0 and y != 0 and z != 0 and x != length - 1 and y != width - 1 and z != height - 1
    }

    # Smooth array levels based on neighbor counts.
    for z, y, x in neighbors:
        array[z, y, x] = sum(array[z + k, y + j, x + i] == 7 for k, j, i in offsets) + 1

    return array


def make_mesh_object(array: np.ndarray) -> str:
    verts, faces, normals, _ = measure.marching_cubes(array, level=3, allow_degenerate=False)
    mesh = ""
    faces = faces + 1

    for item in verts:
        mesh += f"v {item[0]} {item[1]} {item[2]}\n"

    for item in normals:
        mesh += f"vn {item[0]} {item[1]} {item[2]}\n"

    for item in faces:
        mesh += f"f {item[2]}//{item[2]} {item[1]}//{item[1]} {item[0]}//{item[0]}\n"

    return mesh



from typing import Optional

from prefect import task


@task
def get_location_voxels(location: dict, region: Optional[str] = None) -> list[tuple[int, int, int]]:
    voxels = [
        voxel
        for loc in location["location"]
        for voxel in loc["voxels"]
        if not region or loc["region"] == region
    ]
    return voxels



from typing import Optional, Union

import pandas as pd
from prefect import task


@task
def convert_model_units(
    data: pd.DataFrame, ds: float, dt: float, regions: Optional[Union[list[str], str]] = None
) -> None:
    data["time"] = round(dt * data["TICK"], 2)

    if "NUM_VOXELS" in data.columns:
        data["volume"] = ds * ds * ds * data["NUM_VOXELS"]

    if "MAX_Z" in data.columns and "MIN_Z" in data.columns:
        data["height"] = ds * (data["MAX_Z"] - data["MIN_Z"] + 1)

    if "CX" in data.columns:
        data["cx"] = ds * data["CX"]

    if "CY" in data.columns:
        data["cy"] = ds * data["CY"]

    if "CZ" in data.columns:
        data["cz"] = ds * data["CZ"]

    if regions is None:
        return

    if isinstance(regions, str):
        regions = [regions]

    for region in regions:
        if region == "DEFAULT":
            continue

        data[f"volume.{region}"] = ds * ds * ds * data[f"NUM_VOXELS.{region}"]
        data[f"height.{region}"] = ds * (data[f"MAX_Z.{region}"] - data[f"MIN_Z.{region}"])



from typing import Optional

import numpy as np
import pandas as pd
from prefect import task


@task
def merge_region_samples(
    samples: dict[str, pd.DataFrame], margins: tuple[int, int, int]
) -> pd.DataFrame:
    default_samples = samples["DEFAULT"]
    all_samples = tranform_sample_coordinates(default_samples, margins)

    regions = [key for key in samples.keys() if key != "DEFAULT"]
    all_region_samples = []

    for region in regions:
        region_samples = tranform_sample_coordinates(samples[region], margins, default_samples)
        region_samples["region"] = region
        all_region_samples.append(region_samples)

    if len(all_region_samples) > 0:
        all_samples = all_samples.merge(
            pd.concat(all_region_samples), on=["id", "x", "y", "z"], how="left"
        )
        all_samples["region"].fillna("DEFAULT", inplace=True)

    valid_samples = filter_valid_samples(all_samples)

    return valid_samples


def tranform_sample_coordinates(
    samples: pd.DataFrame,
    margins: tuple[int, int, int],
    reference: Optional[pd.DataFrame] = None,
) -> pd.DataFrame:
    """
    Transforms samples into centered coordinates.

    Parameters
    ----------
    samples
        Sample cell ids and coordinates.
    margins
        Margin size in x, y, and z directions.
    reference
        Reference samples used to calculate transformation.

    Returns
    -------
    :
        Transformed sample cell ids and coordinates.
    """
    if reference is None:
        reference = samples

    minimums = (min(reference.x), min(reference.y), min(reference.z))
    offsets = np.subtract(margins, minimums) + 1

    coordinates = samples[["x", "y", "z"]].values + offsets
    coordinates = coordinates.astype("int64")

    transformed_samples = pd.DataFrame(coordinates, columns=["x", "y", "z"])
    transformed_samples.insert(0, "id", samples["id"])

    return transformed_samples


def filter_valid_samples(samples: pd.DataFrame) -> pd.DataFrame:
    """
    Filters samples for valid cell ids.

    Filter conditions include:

    - Each cell must have at least one sample assigned to each specified region

    Parameters
    ----------
    samples
        Sample cell ids and coordinates.

    Returns
    -------
    :
        Valid sample cell ids and coordinates.
    """
    if "region" in samples.columns:
        num_regions = len(samples.region.unique())
        samples = samples.groupby("id").filter(lambda x: len(x.region.unique()) == num_regions)

    return samples.reset_index(drop=True)



from typing import Optional

import pandas as pd
from prefect import task


@task
def convert_to_locations_file(samples: pd.DataFrame) -> list[dict]:
    locations: list[dict] = []
    samples_by_id = samples.groupby("id")

    for i, (_, group) in enumerate(samples_by_id):
        locations.append(convert_to_location(i + 1, group))

    return locations


def convert_to_location(cell_id: int, samples: pd.DataFrame) -> dict:
    """
    Convert samples to ARCADE .LOCATIONS json format.

    Parameters
    ----------
    cell_id
        Unique cell id.
    samples
        Sample cell ids and coordinates.

    Returns
    -------
    :
        Dictionary in ARCADE .LOCATIONS json format.
    """
    center = get_center_voxel(samples)

    if "region" in samples.columns:
        voxels = [
            {"region": region, "voxels": get_location_voxels(samples, region)}
            for region in samples["region"].unique()
        ]
    else:
        voxels = [{"region": "UNDEFINED", "voxels": get_location_voxels(samples)}]

    location = {
        "id": cell_id,
        "center": center,
        "location": voxels,
    }
    return location


def get_center_voxel(samples: pd.DataFrame) -> tuple[int, int, int]:
    """
    Gets coordinates of center voxel of samples.

    Parameters
    ----------
    samples
        Sample cell ids and coordinates.

    Returns
    -------
    :
        Center voxel.
    """
    center_x = int(samples["x"].mean())
    center_y = int(samples["y"].mean())
    center_z = int(samples["z"].mean())
    center = (center_x, center_y, center_z)
    return center


def get_location_voxels(
    samples: pd.DataFrame, region: Optional[str] = None
) -> list[tuple[int, int, int]]:
    """
    Get list of voxel coordinates from samples dataframe.

    Parameters
    ----------
    samples
        Sample cell ids and coordinates.
    region
        Region key.

    Returns
    -------
    :
        List of voxel coordinates.
    """
    if region is not None:
        region_samples = samples[samples["region"] == region]
        voxels_x = region_samples["x"]
        voxels_y = region_samples["y"]
        voxels_z = region_samples["z"]
    else:
        voxels_x = samples["x"]
        voxels_y = samples["y"]
        voxels_z = samples["z"]

    return list(zip(voxels_x, voxels_y, voxels_z))



import xml.etree.ElementTree as ET
from typing import Optional

import numpy as np
import pandas as pd
from prefect import task


@task
def generate_setup_file(
    samples: pd.DataFrame, margins: tuple[int, int, int], potts_terms: list[str]
) -> str:
    init = len(samples["id"].unique())
    bounds = calculate_sample_bounds(samples, margins)
    regions = samples["regions"].unique() if "regions" in samples else None
    setup = make_setup_file(init, bounds, potts_terms, regions)
    return setup


def calculate_sample_bounds(
    samples: pd.DataFrame, margins: tuple[int, int, int]
) -> tuple[int, int, int]:
    """
    Calculate transformed sample bounds including margin.

    Parameters
    ----------
    samples
        Sample cell ids and coordinates.
    margins
        Margin size in x, y, and z directions.

    Returns
    -------
    :
        Bounds in x, y, and z directions.
    """
    mins = (min(samples.x), min(samples.y), min(samples.z))
    maxs = (max(samples.x), max(samples.y), max(samples.z))

    bound_x, bound_y, bound_z = np.subtract(maxs, mins) + np.multiply(2, margins) + 3
    bounds = (bound_x, bound_y, bound_z)

    return bounds


def make_setup_file(
    init: int,
    bounds: tuple[int, int, int],
    terms: list[str],
    regions: Optional[list[str]] = None,
) -> str:
    """
    Create ARCADE setup file.

    Parameters
    ----------
    init
        Number of initial cells.
    bounds
        Bounds in x, y, and z directions.
    regions
        List of regions.
    terms
        List of Potts Hamiltonian terms for setup file.

    Returns
    -------
        Contents of ARCADE setup file.
    """
    root = ET.fromstring("<set></set>")
    series = ET.SubElement(
        root,
        "series",
        {
            "name": "ARCADE",
            "interval": "1",
            "start": "0",
            "end": "0",
            "dt": "1",
            "ds": "1",
            "ticks": "1",
            "length": str(int(bounds[0])),
            "width": str(int(bounds[1])),
            "height": str(int(bounds[2])),
        },
    )

    potts = ET.SubElement(series, "potts")
    for term in terms:
        ET.SubElement(potts, "potts.term", {"id": term})

    agents = ET.SubElement(series, "agents")
    populations = ET.SubElement(agents, "populations")
    population = ET.SubElement(populations, "population", {"id": "X", "init": str(init)})

    if regions is not None:
        for region in regions:
            ET.SubElement(population, "population.region", {"id": region})

    ET.indent(root, space="    ", level=0)
    return ET.tostring(root, encoding="unicode")



from .convert_to_cells_file import convert_to_cells_file
from .convert_to_locations_file import convert_to_locations_file
from .generate_setup_file import generate_setup_file
from .group_template_conditions import group_template_conditions
from .merge_region_samples import merge_region_samples



import pandas as pd
from prefect import task


@task
def convert_to_cells_file(
    samples: pd.DataFrame,
    reference: pd.DataFrame,
    volume_distributions: dict[str, tuple[float, float]],
    height_distributions: dict[str, tuple[float, float]],
    critical_volume_distributions: dict[str, tuple[float, float]],
    critical_height_distributions: dict[str, tuple[float, float]],
    state_thresholds: dict[str, float],
) -> list[dict]:
    cells: list[dict] = []
    samples_by_id = samples.groupby("id")

    for i, (cell_id, group) in enumerate(samples_by_id):
        cell_reference = filter_cell_reference(cell_id, reference)
        cells.append(
            convert_to_cell(
                i + 1,
                group,
                cell_reference,
                volume_distributions,
                height_distributions,
                critical_volume_distributions,
                critical_height_distributions,
                state_thresholds,
            )
        )

    return cells


def convert_to_cell(
    cell_id: int,
    samples: pd.DataFrame,
    reference: dict,
    volume_distributions: dict[str, tuple[float, float]],
    height_distributions: dict[str, tuple[float, float]],
    critical_volume_distributions: dict[str, tuple[float, float]],
    critical_height_distributions: dict[str, tuple[float, float]],
    state_thresholds: dict[str, float],
) -> dict:
    """
    Convert samples to ARCADE .CELLS json format.

    Parameters
    ----------
    cell_id
        Unique cell id.
    samples
        Sample cell ids and coordinates.
    reference
        Reference data for conversion.
    volume_distribution
        Average and standard deviation of volume distributions.
    height_distribution
        Average and standard deviation of height distributions.
    critical_volume_distribution
        Average and standard deviation of critical volume distributions.
    critical_height_distribution
        Average and standard deviation of critical height distributions.

    Returns
    -------
    :
        Dictionary in ARCADE .CELLS json format.
    """
    volume = len(samples)
    height = samples.z.max() - samples.z.min()

    critical_volume = convert_value_distribution(
        (reference["volume"] if "volume" in reference else volume),
        volume_distributions["DEFAULT"],
        critical_volume_distributions["DEFAULT"],
    )

    critical_height = convert_value_distribution(
        (reference["height"] if "height" in reference else height),
        height_distributions["DEFAULT"],
        critical_height_distributions["DEFAULT"],
    )

    state = get_cell_state(volume, critical_volume, state_thresholds)

    cell = {
        "id": cell_id,
        "parent": 0,
        "pop": 1,
        "age": 0,
        "divisions": 0,
        "state": state.split("_")[0],
        "phase": state,
        "voxels": volume,
        "criticals": [critical_volume, critical_height],
    }

    if "region" in samples.columns:
        regions = [
            convert_to_cell_region(
                region,
                region_samples,
                reference,
                volume_distributions,
                height_distributions,
                critical_volume_distributions,
                critical_height_distributions,
            )
            for region, region_samples in samples.groupby("region")
        ]
        cell.update({"regions": regions})

    return cell


def convert_to_cell_region(
    region: str,
    region_samples: pd.DataFrame,
    reference: dict,
    volume_distributions: dict[str, tuple[float, float]],
    height_distributions: dict[str, tuple[float, float]],
    critical_volume_distributions: dict[str, tuple[float, float]],
    critical_height_distributions: dict[str, tuple[float, float]],
) -> dict:
    region_volume = len(region_samples)
    region_height = region_samples.z.max() - region_samples.z.min()

    region_critical_volume = convert_value_distribution(
        (reference[f"volume.{region}"] if f"volume.{region}" in reference else region_volume),
        volume_distributions[region],
        critical_volume_distributions[region],
    )

    region_critical_height = convert_value_distribution(
        (reference[f"height.{region}"] if f"height.{region}" in reference else region_height),
        height_distributions[region],
        critical_height_distributions[region],
    )

    return {
        "region": region,
        "voxels": len(region_samples),
        "criticals": [region_critical_volume, region_critical_height],
    }


def get_cell_state(
    volume: float,
    critical_volume: float,
    threshold_fractions: dict[str, float],
) -> str:
    """
    Estimates cell state based on cell volume.

    The threshold fractions dictionary defines the monotonic thresholds
    between different cell states.
    For a given volume v, critical volume V, and states X1, X2, ..., XN with
    corresponding, monotonic threshold fractions f1, f2, ..., fN, a cell is
    assigned state Xi such that [f(i - 1) * V] <= v < [fi * V].
    Cells with v < f1 * V are assigned state X1.
    Cells with v > fN * V are assigned state XN.

    Parameters
    ----------
    volume
        Current cell volume.
    critical_volume
        Critical cell volume.
    threshold_fractions
        Critical volume fractions defining threshold between states.

    Returns
    -------
    :
        Cell state.
    """
    thresholds = [fraction * critical_volume for fraction in threshold_fractions.values()]
    states = list(threshold_fractions.keys())

    index = next((ind for ind, thresh in enumerate(thresholds) if thresh > volume), -1)
    return states[index]


def convert_value_distribution(
    value: float,
    source_distribution: tuple[float, float],
    target_distribution: tuple[float, float],
) -> float:
    """
    Estimates target value based on source value and source and target distributions.

    Parameters
    ----------
    value
        Source value.
    source_distribution
        Average and standard deviation of source value distribution.
    target_distribution
        Average and standard deviation of target value distribution.

    Returns
    -------
    :
        Estimated critical value.
    """
    source_avg, source_std = source_distribution
    target_avg, target_std = target_distribution
    z_scored_value = (value - source_avg) / source_std
    converted_value = z_scored_value * target_std + target_avg
    return converted_value


def filter_cell_reference(cell_id: int, reference: pd.DataFrame) -> dict:
    """
    Filters reference data for given cell id.

    Parameters
    ----------
    cell_id
        Unique cell id.
    reference
        Reference data for conversion.

    Returns
    -------
    :
        Reference data for given cell id.
    """
    cell_reference = reference[reference["ID"] == cell_id].squeeze()
    cell_reference = cell_reference.to_dict() if not cell_reference.empty else {}
    return cell_reference



from .convert_model_units import convert_model_units
from .convert_to_images import convert_to_images
from .convert_to_meshes import convert_to_meshes
from .convert_to_simularium import convert_to_simularium
from .extract_tick_json import extract_tick_json
from .get_location_voxels import get_location_voxels
from .merge_parsed_results import merge_parsed_results
from .parse_cells_file import parse_cells_file
from .parse_locations_file import parse_locations_file
